/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.util;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.intoJ.riwd.multirefac.Activator;

/**
 * Logging.
 * 
 * @version: $Date: 2007/09/09 19:40:25 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class ILog {

	private static final String PLUGIN_NAME = "InferType";

	public static int LOG_ADMIN = 0;
	public static int LOG_GATHER = 1;
	public static int LOG_STEPS = 2;
	public static int LOG_MANIPULATE = 3;
	public static int LOG_AST = 4;
	public static int LOG_EVAL = 5;

	private static boolean[] loggingEnabled = new boolean[] { false, false, false, false, false, false };
	private static String[] partName = new String[] { "AD", "GA", "ST", "MA", "AS", "EV" };

	static {
		loggingEnabled[LOG_ADMIN] = "true".equalsIgnoreCase(Platform.getDebugOption("org.intoJ.inferType/trace/administrative"));
		loggingEnabled[LOG_GATHER] = "true".equalsIgnoreCase(Platform.getDebugOption("org.intoJ.inferType/trace/infogathering"));
		loggingEnabled[LOG_STEPS] = "true".equalsIgnoreCase(Platform.getDebugOption("org.intoJ.inferType/trace/typeinference"));
		loggingEnabled[LOG_MANIPULATE] = "true".equalsIgnoreCase(Platform.getDebugOption("org.intoJ.inferType/trace/codemanipulation"));
		loggingEnabled[LOG_AST] = "true".equalsIgnoreCase(Platform.getDebugOption("org.intoJ.inferType/trace/astoperations_detailed"));
		loggingEnabled[LOG_EVAL] = "true".equalsIgnoreCase(Platform.getDebugOption("org.intoJ.inferType/trace/evaluation"));

		informUserAboutLogging(PLUGIN_NAME + " (Administrative)", loggingEnabled[LOG_ADMIN]);
		informUserAboutLogging(PLUGIN_NAME + " (InfoGathering)", loggingEnabled[LOG_GATHER]);
		informUserAboutLogging(PLUGIN_NAME + " (TypeInference)", loggingEnabled[LOG_STEPS]);
		informUserAboutLogging(PLUGIN_NAME + " (Manipulation)", loggingEnabled[LOG_MANIPULATE]);
		informUserAboutLogging(PLUGIN_NAME + " (AST)", loggingEnabled[LOG_AST]);
		informUserAboutLogging(PLUGIN_NAME + " (EVAL)", loggingEnabled[LOG_EVAL]);
	}

	/**
	 * Log an error
	 * @param message String
	 * @param t Throwable
	 */
	public static void logError(String message, Throwable t) {

		Activator.getDefault().getLog().log(new Status(IStatus.ERROR, Activator.getDefault().getBundle().getSymbolicName(), 0, message, t));
	}

	/**
	 * Log stuff 
	 * @param level int
	 * @param msg String
	 */
	public static void log(int part, int level, String msg) {

		if (loggingEnabled[part]) {
			StringBuffer b = new StringBuffer();
			for (int i = 0; i < level; i++) {
				b.append("  ");
			}

			System.out.println(PLUGIN_NAME + " :: " + partName[part] + " :: " + b.toString() + msg);
		}
	}

	private static void informUserAboutLogging(String which, boolean loggingEnabled) {

		if (loggingEnabled)
			System.out.println("Enabled logging for " + which);
		else
			System.out.println("No tracing logging for " + which);
	}

}