/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.util;

import java.util.ArrayList;
import java.util.Collection;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.search.SearchMatch;
import org.eclipse.jdt.core.search.SearchRequestor;

/**
 * This collector is being called back from the search engine to collect the
 * results.
 * 
 * We are primarily interested in compilation units here.
 * 
 * @version: $Date: 2007/09/09 19:40:25 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 *  
 */
public class SearchResultCollector extends SearchRequestor {

	private Collection compilationUnits;
	private Collection compilationUnitsHandleIdentifiers;

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.core.search.IJavaSearchResultCollector#aboutToStart()
	 */
	public void beginReporting() {

		compilationUnits = new ArrayList();
		compilationUnitsHandleIdentifiers = new ArrayList();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.core.search.IJavaSearchResultCollector#accept(org.eclipse.core.resources.IResource,
	 *      int, int, org.eclipse.jdt.core.IJavaElement, int)
	 */
	public void acceptSearchMatch(SearchMatch match) throws CoreException {

		IJavaElement enclosingElement = (IJavaElement) match.getElement();

		if (enclosingElement != null) {
			ICompilationUnit unit = (ICompilationUnit) enclosingElement.getAncestor(IJavaElement.COMPILATION_UNIT);

			if (unit != null) {
				if (!compilationUnitsHandleIdentifiers.contains(unit.getHandleIdentifier())) {
					// Comparing the compilation units itself doesn't work,
					// So we need to compare the handles, which are unique.
					compilationUnits.add(unit);
					compilationUnitsHandleIdentifiers.add(unit.getHandleIdentifier());
				}
			}
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.core.search.IJavaSearchResultCollector#done()
	 */
	public void done() {

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.core.search.IJavaSearchResultCollector#getProgressMonitor()
	 */
	public IProgressMonitor getProgressMonitor() {

		return null;
	}

	/**
	 * Returns the compilation units.
	 * 
	 * @return Collection of ICompilationUnit
	 */
	public Collection getCompilationUnits() {

		return compilationUnits;
	}

}