package org.intoj.riwd;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.BindingKey;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.ITypeParameter;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.EnumDeclaration;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.ParameterizedType;
import org.eclipse.jdt.core.dom.SuperConstructorInvocation;
import org.eclipse.jdt.core.dom.SuperMethodInvocation;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.jdt.core.dom.rewrite.ListRewrite;
import org.eclipse.jdt.internal.corext.codemanipulation.StubUtility;
import org.eclipse.jdt.internal.corext.dom.Bindings;
import org.eclipse.jdt.internal.corext.refactoring.Checks;
import org.eclipse.jdt.internal.corext.refactoring.RefactoringCoreMessages;
import org.eclipse.jdt.internal.corext.refactoring.structure.ASTNodeSearchUtil;
import org.eclipse.jdt.internal.corext.refactoring.structure.CompilationUnitRewrite;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2;
import org.eclipse.jdt.internal.corext.refactoring.util.ResourceUtil;
import org.eclipse.jdt.internal.corext.refactoring.util.TextEditBasedChangeManager;
import org.eclipse.ltk.core.refactoring.CategorizedTextEditGroup;
import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.CompositeChange;
import org.eclipse.ltk.core.refactoring.GroupCategory;
import org.eclipse.ltk.core.refactoring.GroupCategorySet;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.eclipse.ltk.core.refactoring.TextEditBasedChange;
import org.intoJ.inferType3.InferTypeRefactoring;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.variables.BindingKeyProvider;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutabilityReason;
import org.intoJ.inferType3.internal.constraintModel.variables.SuperVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ThisVariable;
import org.intoJ.inferType3.internal.manipulation.CodeGenerator;
import org.intoJ.inferType3.internal.manipulation.GenerationUtil;
import org.intoJ.inferType3.internal.typeCalc.AccessSet;
import org.intoJ.inferType3.internal.typeCalc.AccessSetCalculation;
import org.intoJ.inferType3.internal.typeCalc.BindingSet;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ParameterizedTypeUtil;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;
import org.intoJ.inferType3.internal.util.Cache;
import org.intoJ.inferType3.internal.util.InternalCalculationException;
import org.intoj.riwd.RiwdRefactoring.Status;

public class Manipulation {

	private static final String FIELD_BASE_NAME = "delegatee";
	private static final String CLASS_BASE_NAME = "Delegatee";
    private static String SUPER_PREFIX = "_super_";

    private String delegateeFieldName;

	private static final String ID_RIWD = "org.intoJ.riwd";

    private static final GroupCategorySet SET_RIWD = new GroupCategorySet(new GroupCategory(ID_RIWD, "Replace Inheritance With Delegation", "Replace Inheritance With Delegation"));

    private TextEditBasedChangeManager manager = new TextEditBasedChangeManager();

    private final TypeHierarchy typeHierarchy;
	private ExistingTCType newSuperclass;
	private final CompilationUnitRewrite sourceRewrite;
	private final SuperCallVisitor superCallVisitor;
	private boolean performForwarding;
	private String delegateeClassName;

	private Set<IMethodBinding> superMethodsNeeded = new HashSet<IMethodBinding>();
	private Set<IMethodBinding> wideningMethodsNeeded = new HashSet<IMethodBinding>();
	private final Status commonStatus;
	
	private final Stats stats = new Stats();
	
    
    public Manipulation(TypeHierarchy typeHierarchy, CompilationUnitRewrite sourceRewrite, SuperCallVisitor superCallVisitor, Status commonStatus) {
		this.typeHierarchy = typeHierarchy;
		this.sourceRewrite = sourceRewrite;
		this.superCallVisitor = superCallVisitor;
		this.commonStatus = commonStatus;
		
	}
	
    public final void calculateChanges(final IProgressMonitor monitor, final RefactoringStatus status, Object validationContext, boolean performForwarding, TType newTSuperclass, Set<TType> requiredInterfaces, InferTypeConstraintsModel model) throws JavaModelException, CoreException, InternalCalculationException {        
        Assert.isNotNull(status);
        Assert.isNotNull(monitor);        
        this.newSuperclass = typeHierarchy.getType(newTSuperclass);
        this.performForwarding = performForwarding;
        
        stats.newSupertype = newSuperclass != null;
        
        try {
            monitor.beginTask("", 85);
            monitor.setTaskName(InferTypeRefactoring.MSG_CREATING_CHANGES);

            final IJavaProject project = typeHierarchy.getOriginalSelectedType().getBaseTypeBinding().getJavaElement().getJavaProject();
            
            findFieldName();
            findDelegateeClassName();

            monitor.worked(10);

            ExistingTCType type = typeHierarchy.getOriginalSelectedType();
            
    		IType iType = type.getIType();
                        
            BindingSet usedMethods = calculateUsedMethods(model);
            
            updateRequiredInterfaces(requiredInterfaces, type);
            
            
            try {
                final AbstractTypeDeclaration declaration = ASTNodeSearchUtil.getAbstractTypeDeclarationNode(iType, sourceRewrite.getRoot());
                if (declaration != null) {            	
                    changeTypeSignature(declaration, requiredInterfaces, status);                
                    createField(type, declaration);                
                    changeSuperConstructors(type, status);
                    addSuperConstructors(type, status);
                    changeSuperMethodInvocations(type, status);
                    
                    Map<String, Integer> addedMethods = new HashMap<String, Integer>();
                    Map<String, IMethodBinding> overriddenMethods = null;
                    if (!performForwarding) {
                    	overriddenMethods = getOverridenMethods(type);
                    	addMethodsOverridenInSubclasses(type, overriddenMethods, addedMethods, declaration, project);
                    }
                    addUsedMethods(type, addedMethods, requiredInterfaces, usedMethods, declaration, project, model, status);
                    if (!performForwarding) {
                    	addDelegateeClass(type, overriddenMethods, declaration, project);
                    }
                    
                    calculateStats(usedMethods, addedMethods, requiredInterfaces);
                }
	            manager.manage(sourceRewrite.getCu(), sourceRewrite.createChange());
	            
	            monitor.worked(10);

            } catch (Exception e) {
                System.err.println("Could not modify class: " + e);
                status.addFatalError("Could not modify class: " + e);
                e.printStackTrace();
            }
            if (!status.hasFatalError()) {
                final RefactoringStatus validation = Checks.validateModifiesFiles(ResourceUtil.getFiles(manager.getAllCompilationUnits()), validationContext);
                if (!validation.isOK()) {
                    status.merge(validation);
                }
            }
            

        } finally {
            monitor.done();
        }
    }

	private void calculateStats(BindingSet usedMethods,
			Map<String, Integer> addedMethods, Set<TType> requiredInterfaces) {
		
		stats.newInterfaces = requiredInterfaces.size();
			
		ExistingTCType type = typeHierarchy.getOriginalSelectedType();
		
		BindingSet allMethods = type.getAllMethods();
		for (BindingWrapper bindingWrapper : allMethods) {			
			IMethodBinding binding = bindingWrapper.getMethodBinding();
			int modifiers = binding.getModifiers();
			if (!AccessSet.isObjectMethod(binding) && !Modifier.isPrivate(modifiers)) {
				if (Modifier.isStatic(modifiers)) {
					stats.nonPrivateStaticMethodsBefore++;
				} else {				
					stats.nonPrivateMethodsBefore++;
				}
			}
		}
		
		BindingSet declaredMethods = type.getDeclaredMethods();
		for (BindingWrapper bindingWrapper : declaredMethods) {			
			IMethodBinding binding = bindingWrapper.getMethodBinding();
			int modifiers = binding.getModifiers();
			if (!AccessSet.isObjectMethod(binding) && !Modifier.isPrivate(modifiers)) {
				if (Bindings.findMethodInHierarchy(type.getBaseTypeBinding().getSuperclass(), 
						binding.getName(), binding.getParameterTypes()) == null) {
					if (Modifier.isStatic(modifiers)) {
						stats.newlyDeclaredNonPrivateStaticMethodsBefore++;
					} else {				
						stats.newlyDeclaredNonPrivateMethodsBefore++;
					}					
				}
			}
		}		

		stats.onlyInheritedMethodsBefore = stats.nonPrivateMethodsBefore - stats.newlyDeclaredNonPrivateMethodsBefore;
		stats.onlyInheritedStaticMethodsBefore = stats.nonPrivateStaticMethodsBefore - stats.newlyDeclaredNonPrivateStaticMethodsBefore;
		
		
		Set<String> baseMethodSigs = new HashSet<String>();
		Set<String> baseStaticMethodSigs = new HashSet<String>();

		if (newSuperclass != null) {
			for (BindingWrapper bindingWrapper : newSuperclass.getAllMethods()) {			
				IMethodBinding binding = bindingWrapper.getMethodBinding();
				int modifiers = binding.getModifiers();
				if (!AccessSet.isObjectMethod(binding) && !Modifier.isPrivate(modifiers)) {
			        String sig = binding.getName() + new BindingKey(binding.getKey()).toSignature();
					
					if (Modifier.isStatic(modifiers)) {
						baseStaticMethodSigs.add(sig);
					} else {				
						baseMethodSigs.add(sig);
					}
				}
			}
		}
		
		for (BindingWrapper bindingWrapper : type.getDeclaredMethods()) {			
			IMethodBinding binding = bindingWrapper.getMethodBinding();
			int modifiers = binding.getModifiers();
			if (!AccessSet.isObjectMethod(binding) && !Modifier.isPrivate(modifiers)) {
		        String sig = binding.getName() + new BindingKey(binding.getKey()).toSignature();

		        if (Modifier.isStatic(modifiers)) {
					baseStaticMethodSigs.add(sig);
				} else {				
					baseMethodSigs.add(sig);
				}
			}
		}
		
		for (Map.Entry<String, Integer> entry : addedMethods.entrySet()) {
			int modifier = entry.getValue().intValue();
			if (!Modifier.isPrivate(modifier)) {
				if (Modifier.isStatic(modifier)) {
					baseStaticMethodSigs.add(entry.getKey());				
				} else {
					baseMethodSigs.add(entry.getKey());				
				}
			}
		}
		
		stats.nonPrivateMethodsAfter = baseMethodSigs.size();
		stats.nonPrivateStaticMethodsAfter = baseStaticMethodSigs.size();
		
		Set<String> usedMethodSigs = new HashSet<String>();
		for (BindingWrapper bindingWrapper : usedMethods) {
	        IMethodBinding binding = bindingWrapper.getMethodBinding();
			String sig = binding.getName() + new BindingKey(binding.getKey()).toSignature();
			usedMethodSigs.add(sig);
		}
		
		for (String sig : baseMethodSigs) {
			if (!usedMethodSigs.contains(sig)) {
				stats.deadMethods++;
			}
		}

		for (String sig : baseStaticMethodSigs) {
			if (!usedMethodSigs.contains(sig)) {
				stats.deadMethods++;
			}
		}
		
	}

	private void addDelegateeClass(ExistingTCType type,
			Map<String, IMethodBinding> overriddenMethods, AbstractTypeDeclaration declaration, IJavaProject javaProject) throws JavaModelException, InternalCalculationException {

		ASTRewrite astRewrite = sourceRewrite.getASTRewrite();

		StringBuilder content = new StringBuilder();
		ImportRewrite importRewrite = sourceRewrite.getImportRewrite();
		
		String lineDelimiter = StubUtility.getLineDelimiterUsed(type.getIType());
		final String delimiter= lineDelimiter + "    ";

		content.append(lineDelimiter).append("protected class ").append(delegateeClassName);
        StringBuffer methodTypeParameterList = GenerationUtil.getTypeParameterList(type.getBaseTypeBinding().getTypeParameters(), importRewrite);
        if (methodTypeParameterList.length() != 0) {
        	content.append("<").append(methodTypeParameterList).append(">");
        }		
        content.append(" extends ");
		content.append(importRewrite.addImport(type.getBaseTypeBinding().getSuperclass()));
		content.append(" {").append(delimiter).append(delimiter);
		if (superCallVisitor.getConstructorWithoutSuperList().size() > 0 && superCallVisitor.getSuperConstructorInvocationList().size() > 0) {
			content.append("public ").append(delegateeClassName).append("() { ").append(delimiter);
			content.append("}").append(delimiter);
		}
		for (SuperConstructorInvocation superConstructorInvocation : superCallVisitor.getSuperConstructorInvocationList()) {
			content.append(delimiter);			
			addMethodContent(superConstructorInvocation.resolveConstructorBinding(), 
					type, null, content,
					importRewrite, delimiter, "super", delegateeClassName);
			content.append(delimiter);			
		}

		Set<String> addedMethods = new HashSet<String>();
		for (IMethodBinding binding : overriddenMethods.values()) {
			addedMethods.add(binding.getName() + new BindingKey(binding.getKey()).toSignature());
			content.append(delimiter);		
			stats.reverseDelegatingMethods++;
			addMethodContent(binding, type, null, content, importRewrite, 
					delimiter, type.getSimpleName() + ".this." + binding.getName(), binding.getName());					
			content.append(delimiter);			
		}
		
		for (IMethodBinding binding  : wideningMethodsNeeded) {
			if (addedMethods.add(binding.getName() + new BindingKey(binding.getKey()).toSignature())) {
				content.append(delimiter);			
				addMethodContent(binding, type, null, content, importRewrite, 
						delimiter, "super." + binding.getName(), binding.getName());					
				content.append(delimiter);							
			}
		}

		for (IMethodBinding binding : superMethodsNeeded) {
			content.append(delimiter);			
			addMethodContent(binding, type, null, content, importRewrite, 
					delimiter, "super." + binding.getName(), SUPER_PREFIX + binding.getName());					
			content.append(delimiter);			
		}
		content.append(lineDelimiter);
		content.append("}").append(lineDelimiter);
				
        // CodeFormatterUtil.format(CodeFormatter.K_CLASS_BODY_DECLARATIONS, contents.toString(), 0, null, delimiter, javaProject)
		final TypeDeclaration typeDeclaration = (TypeDeclaration) astRewrite.createStringPlaceholder(
				content.toString(), ASTNode.TYPE_DECLARATION);
		
		ListRewrite listRewriter = astRewrite.getListRewrite(declaration, declaration.getBodyDeclarationsProperty());
		listRewriter.insertLast(typeDeclaration, sourceRewrite.createCategorizedGroupDescription("Add delegation method", SET_RIWD));
	}

	private Map<String, IMethodBinding> getOverridenMethods(ExistingTCType type) {
		HashMap<String, IMethodBinding> ret = new HashMap<String, IMethodBinding>();
		
		addOverriddenMethods(type.getSuperType(), type, ret);
		for (ExistingTCType subType : typeHierarchy.getSubTypes()) {
			addOverriddenMethods(type.getSuperType(), subType, ret);
		}
		return ret;
	}

	private void addOverriddenMethods(TCType baseType, TCType checkType,
			HashMap<String, IMethodBinding> map) {
		for (IMethodBinding binding : checkType.getBaseTypeBinding().getDeclaredMethods()) {
			if (Bindings.findMethodInHierarchy(baseType.getBaseTypeBinding(), binding.getName(), binding.getParameterTypes()) != null) {
		        String key = binding.getName() + new BindingKey(binding.getKey()).toSignature();
		        if (!map.containsKey(key)) {
		        	map.put(key, binding);
		        }
			}
		}
	}

	private void findDelegateeClassName() {
		if (performForwarding) {
			ImportRewrite importRewrite = sourceRewrite.getImportRewrite();
			delegateeClassName = importRewrite.addImport(typeHierarchy.getOriginalSelectedType().getSuperType().getBaseTypeBinding());
		} else {
			Set<String> declaredTypeNames = new HashSet<String>();
			for (ITypeBinding typeBinding : typeHierarchy.getOriginalSelectedType().getBaseTypeBinding().getDeclaredTypes()) {
				declaredTypeNames.add(typeBinding.getName());
			}
			
			delegateeClassName = CLASS_BASE_NAME;
			int i=2;
			while (declaredTypeNames.contains(delegateeClassName)) {
				delegateeClassName = CLASS_BASE_NAME + (i++);
			}			
		}
	}

	private void updateRequiredInterfaces(Set<TType> requiredInterfaces,
			ExistingTCType type) {
		Set<TCType> missingInferfaces = new HashSet<TCType>();
		
		TCType parent = type.getSuperType();
		while (parent != null && !parent.equals(newSuperclass)) {
			missingInferfaces.addAll(parent.getInterfaces());
			parent = parent.getSuperType();
		}
		
		for (Iterator<TType> it = requiredInterfaces.iterator(); it.hasNext();) {
			TType interf = it.next();
			if (!missingInferfaces.contains(typeHierarchy.getType(interf))) {
				it.remove();
			}
			
		}
	}
    
    private BindingSet calculateUsedMethods(InferTypeConstraintsModel model) throws JavaModelException {
    	
    	BindingSet ret = new BindingSet();
    	Set<TType> searchedTypes = new HashSet<TType>();
    	Set<String> subTypes = new HashSet<String>();
    	for (ExistingTCType existingTCType : typeHierarchy.getSubTypes()) {
    		subTypes.add(existingTCType.getTypeName());
			searchedTypes.add(model.createTType(existingTCType.getBaseTypeBinding()));
		}
    	searchedTypes.add(model.createTType(typeHierarchy.getOriginalSelectedType().getBaseTypeBinding()));
    	
    	for (ConstraintVariable2 cv : model.getConstraintVariables()) {
			TType type = cv.getType();
			if (type != null && searchedTypes.contains(type.getErasure())) {
				ret.addAll(InferTypeConstraintsModel.getAccessSet(cv).getMembers());
			}
    	}
    	removeSubMethods(ret);
    	

//		for (ITypeConstraint2 constraint : model.getTypeConstraints()) {			
//			TType leftType = constraint.getLeft().getType();
//			if (leftType != null && searchedTypes.contains(leftType.getErasure())) {
//				if (constraint.getRight() instanceof BindingKeyProvider) {
//					String bindingKey = ((BindingKeyProvider)constraint.getRight()).getBindingKey();
//					if (bindingKey != null) {
//						BindingWrapper method = typeHierarchy.getMethod(bindingKey);
//						if (method != null && !subTypes.contains(method.getMethodBinding().getDeclaringClass().getQualifiedName())) {
//							ret.add(method);						
//						}
//					}
//				}
//			}
//		}
		return ret;
	}
    
	private void removeSubMethods(BindingSet totalSet) throws JavaModelException {
    	IType originalType = (IType)typeHierarchy.getOriginalSelectedType().getBaseTypeBinding().getJavaElement();
    	ITypeHierarchy superTypes = originalType.newSupertypeHierarchy(null);
    	for (Iterator<BindingWrapper> it = totalSet.iterator(); it.hasNext();) {
    		BindingWrapper wrapper = it.next();
    		ITypeBinding declaringClass = wrapper.getMethodBinding().getDeclaringClass();
			if (!superTypes.contains((IType)declaringClass.getJavaElement())) {
				it.remove();
			}			
		}
	}
    

	private void addSuperConstructors(ExistingTCType type,
			RefactoringStatus status) throws JavaModelException, InternalCalculationException {
    	
		ASTRewrite astRewrite = sourceRewrite.getASTRewrite();
		
		ImportRewrite importRewrite = sourceRewrite.getImportRewrite();
		
        StringBuilder builder = new StringBuilder();
		builder.append(delegateeFieldName);
        builder.append(" = new ").append(delegateeClassName);
		addTypeArguments(builder, type, importRewrite);
        builder.append("();");
		
		for (MethodDeclaration methodDeclaration : superCallVisitor.getConstructorWithoutSuperList()) {
			if (methodDeclaration.getBody() != null) {
				ASTNode delegateeConstruction = astRewrite.createStringPlaceholder(builder.toString(), ASTNode.ASSIGNMENT);
				ListRewrite listRewriter = astRewrite.getListRewrite(methodDeclaration.getBody(), Block.STATEMENTS_PROPERTY);
				listRewriter.insertFirst(delegateeConstruction, sourceRewrite.createCategorizedGroupDescription("Add delegatee constructor", SET_RIWD));
			}
		}
		
	}

	private void findFieldName() {
    	ITypeBinding typeBindinding = typeHierarchy.getOriginalSelectedType().getBaseTypeBinding();
    	ITypeBinding superClassBinding = null;
    	if (newSuperclass != null) {
    		superClassBinding = newSuperclass.getBaseTypeBinding();
    	}
    	
    	delegateeFieldName = FIELD_BASE_NAME;
    	int i=1; 
    	while (containsField(delegateeFieldName, typeBindinding, false) ||
    			containsField(delegateeFieldName, superClassBinding, true)) {
    		delegateeFieldName = FIELD_BASE_NAME + (i++);
    	}
	}

	private static boolean containsField(String delegateeFieldName,
			ITypeBinding typeBinding, boolean superclass) 
	{
		if (typeBinding != null) {
			for (IVariableBinding fieldBinding : typeBinding.getDeclaredFields()) {
				if (fieldBinding.getName().equals(delegateeFieldName)) {
					if (!(superclass && Modifier.isPrivate(fieldBinding.getModifiers()))) {
						return true;
					}
				}
			}
			if (superclass) {
				return containsField(delegateeFieldName, typeBinding.getSuperclass(), true);
			}
		}
		return false;
	}

	/**
     * Returns all changes as a single composite change object.
     * 
     * @return all changes or null if there are none
     * @throws JavaModelException
     */
    public Change getAllChanges() throws JavaModelException {
        CompositeChange compositeChange = new CompositeChange("Replace Inheritance With Delegation");

        final TextEditBasedChange[] changes = manager.getAllChanges();
        if (changes != null && changes.length != 0) {
            compositeChange.addAll(changes);
        }
        
        if (compositeChange.getChildren().length > 0) {
            Cache.changesMade();
            return compositeChange;
        } 
        
        return compositeChange;
    }
    
	private void changeSuperMethodInvocations(ExistingTCType type,
			RefactoringStatus status) throws JavaModelException, InternalCalculationException 
	{		
		ASTRewrite astRewrite = sourceRewrite.getASTRewrite();

		for (SuperMethodInvocation superMethodInvocation : superCallVisitor.getSuperMethodInvocationList()) {
			StringBuilder content = new StringBuilder();
			
			IMethodBinding superBinding = superMethodInvocation.resolveMethodBinding();
			
			MethodDeclaration methodDeclaration = (MethodDeclaration)getParent(superMethodInvocation, ASTNode.METHOD_DECLARATION);
			IMethodBinding declaringBinding = methodDeclaration.resolveBinding();
			content.append(delegateeFieldName).append(".");

			if (!performForwarding && ((IMethod)declaringBinding.getJavaElement()).isSimilar((IMethod)superBinding.getJavaElement())) {
				content.append(SUPER_PREFIX);
				superMethodsNeeded.add(superBinding);
			} else if (!Modifier.isPublic(superBinding.getModifiers())) {
				wideningMethodsNeeded.add(superBinding);
			}
			
			content.append(superBinding.getName()).append("(");
			List arguments = superMethodInvocation.arguments();
			if (arguments.size() > 0) {
				Expression lastExpression = ((Expression)arguments.get(arguments.size() - 1));
				copyContent(content, sourceRewrite, ((Expression)arguments.get(0)).getStartPosition(),
						lastExpression.getStartPosition() + lastExpression.getLength()); 
			}
			content.append(")");
			
			final ASTNode replacement = astRewrite.createStringPlaceholder(
					content.toString(), ASTNode.METHOD_INVOCATION);

			astRewrite.replace(superMethodInvocation, replacement, sourceRewrite.createCategorizedGroupDescription("Change super invocation", SET_RIWD));
			
		}		
	}

	private ASTNode getParent(
			ASTNode node, int nodeType) {
		while (node != null && node.getNodeType() != nodeType) {
			node = node.getParent();
		}
		return node;
	}

	private void changeSuperConstructors(ExistingTCType type,
			RefactoringStatus status) throws JavaModelException, InternalCalculationException 
	{		
		ASTRewrite astRewrite = sourceRewrite.getASTRewrite();

		ImportRewrite importRewrite = sourceRewrite.getImportRewrite();

		for (SuperConstructorInvocation superConstructorInvocation : superCallVisitor.getSuperConstructorInvocationList()) {
			StringBuilder content = new StringBuilder();

			content.append(delegateeFieldName).append(" = new ").append(delegateeClassName);
			addTypeArguments(content, type, importRewrite);
			content.append("(");
			List arguments = superConstructorInvocation.arguments();
			if (arguments.size() > 0) {
				Expression lastExpression = ((Expression)arguments.get(arguments.size() - 1));
				copyContent(content, sourceRewrite, ((Expression)arguments.get(0)).getStartPosition(),
						lastExpression.getStartPosition() + lastExpression.getLength()); 
			}
			content.append(");");
			
			final ASTNode replacement = astRewrite.createStringPlaceholder(
					content.toString(), ASTNode.ASSIGNMENT);

			astRewrite.replace(superConstructorInvocation, replacement, sourceRewrite.createCategorizedGroupDescription("Create delegatee", SET_RIWD));
			
		}		
	}

	private void copyContent(StringBuilder content, CompilationUnitRewrite sourceRewrite, int startPosition, int endPosition) throws JavaModelException {
		String text = sourceRewrite.getCu().getBuffer().getText(startPosition, endPosition - startPosition);
		content.append(text);
		
	}

	private void addTypeArguments(StringBuilder content, ExistingTCType type, ImportRewrite importRewrite) throws JavaModelException, InternalCalculationException {
        ITypeBinding constructorClass = performForwarding ? type.getSuperType().getBaseTypeBinding() : type.getBaseTypeBinding();
        if (constructorClass.isGenericType()) {
        	constructorClass = ParameterizedTypeUtil.getParametrizedTypeBinding(constructorClass.getKey(), type.getBaseTypeBinding());
        }

        ITypeBinding[] typeArguments = performForwarding ? constructorClass.getTypeArguments() : constructorClass.getTypeParameters();
		StringBuffer methodTypeParameterList = GenerationUtil.getTypeParameterList(typeArguments, importRewrite);
        if (methodTypeParameterList.length() != 0) {
        	content.append("<").append(methodTypeParameterList).append(">");
        }
	}
	
	private void addMethodsOverridenInSubclasses(ExistingTCType type,
			Map<String, IMethodBinding> overridenMethods,
			Map<String, Integer> addedMethods,
			AbstractTypeDeclaration declaration, IJavaProject javaProject) throws JavaModelException, InternalCalculationException {

		for (Map.Entry<String, IMethodBinding> entry : overridenMethods.entrySet()) {
			if (!entry.getValue().getDeclaringClass().getErasure().getKey().equals(type.getBaseTypeBinding().getErasure().getKey())) {
				if (!addedMethods.containsKey(entry.getKey())) {
					addedMethods.put(entry.getKey(), entry.getValue().getModifiers());
	        		addMethod(entry.getValue(), type, sourceRewrite, declaration, javaProject, null, true);
	        		stats.superDelegatingMethods++;
	        		superMethodsNeeded.add(entry.getValue());
				}
			}			
		}
	}

	private void addUsedMethods(ExistingTCType type, Map<String, Integer> addedMethods,
			Set<TType> requiredInterfaces, BindingSet usedMethods, AbstractTypeDeclaration declaration, IJavaProject javaProject, InferTypeConstraintsModel model, RefactoringStatus status) throws JavaModelException, InternalCalculationException 
	{

		for (TType interf : requiredInterfaces) {
			for (BindingWrapper bindingWrapper : typeHierarchy.getType(interf).getAllMethods()) {
				IMethodBinding binding = bindingWrapper.getMethodBinding();
		        String sig = binding.getName() + new BindingKey(binding.getKey()).toSignature();
				if (!addedMethods.containsKey(sig)) {
		            if (Bindings.findMethodInType(type.getBaseTypeBinding(), 
		            		binding.getName(), binding.getParameterTypes()) == null) 
		            {
						addedMethods.put(sig, binding.getModifiers());
						stats.delegatingMethods++;
		        		addMethod(binding, type, sourceRewrite, declaration, javaProject, null, false);
		            }
				}
			}
		}
		
        ITypeBinding superBinding = null;
        if (newSuperclass != null) {
        	superBinding = newSuperclass.getBaseTypeBinding();
        } else {
        	superBinding = sourceRewrite.getAST().resolveWellKnownType("java.lang.Object");
        }

        for (BindingWrapper bindingWrapper : usedMethods) {
			IMethodBinding binding = bindingWrapper.getMethodBinding();
	        String sig = binding.getName() + new BindingKey(binding.getKey()).toSignature();
			if (!addedMethods.containsKey(sig)) {
	            if (Bindings.findMethodInType(type.getBaseTypeBinding(), 
	            		binding.getName(), binding.getParameterTypes()) == null) 
	            {	    
					stats.delegatingMethods++;
					if (!Modifier.isPublic(binding.getModifiers())) {
						wideningMethodsNeeded.add(binding);
					}

					IMethodBinding newOverride = Bindings.findMethodInHierarchy(superBinding, 
		            		binding.getName(), binding.getParameterTypes());
	            	if (newOverride != null && Modifier.isFinal(newOverride.getModifiers())) {
	            		commonStatus.addViolation("Refactoring would override final method " + newOverride.getKey(), 
	            				ImmutabilityReason.FINAL_METHOD);
	            	}
	        		int newModifier = addMethod(binding, type, sourceRewrite, declaration, javaProject, 
	        				newOverride == null ? model : null, false);
					addedMethods.put(sig, newModifier);
	            }
			}
		}
	}

	private int addMethod(IMethodBinding binding, ExistingTCType type,
			CompilationUnitRewrite sourceRewrite,
			AbstractTypeDeclaration declaration, IJavaProject javaProject, InferTypeConstraintsModel model, boolean useSuperPrefix) throws JavaModelException, InternalCalculationException {
		ASTRewrite astRewrite = sourceRewrite.getASTRewrite();

		ImportRewrite importRewrite = sourceRewrite.getImportRewrite();
		
		final String delimiter= StubUtility.getLineDelimiterUsed(type.getIType());
		
		StringBuilder invocationPrefix = new StringBuilder();
		if (Modifier.isStatic(binding.getModifiers())) {
			invocationPrefix.append(importRewrite.addImport(type.getSuperType().getTypeName()));			
		} else {
			invocationPrefix.append(delegateeFieldName);
		}
		invocationPrefix.append(".");
		if (useSuperPrefix) {
			invocationPrefix.append(SUPER_PREFIX);
		}
		invocationPrefix.append(binding.getName());

		StringBuilder content = new StringBuilder();
		int newModifier = addMethodContent(binding, type, model, content,
				importRewrite, delimiter, invocationPrefix.toString(), binding.getName());		

        // CodeFormatterUtil.format(CodeFormatter.K_CLASS_BODY_DECLARATIONS, contents.toString(), 0, null, delimiter, javaProject)
		final MethodDeclaration methodDeclaration = (MethodDeclaration) astRewrite.createStringPlaceholder(
				content.toString(), ASTNode.METHOD_DECLARATION);
		
		ListRewrite listRewriter = astRewrite.getListRewrite(declaration, declaration.getBodyDeclarationsProperty());
		listRewriter.insertLast(methodDeclaration, sourceRewrite.createCategorizedGroupDescription("Add delegation method", SET_RIWD));
		
		return newModifier;
	}

	private int addMethodContent(IMethodBinding binding, ExistingTCType type,
			InferTypeConstraintsModel model, 
			StringBuilder content, ImportRewrite importRewrite,
			String delimiter, String invocationPrefix, String methodName)
			throws JavaModelException, InternalCalculationException {

		binding = CodeGenerator.findParametrizedBinding(binding, type.getBaseTypeBinding());
            
        int modifier = getModifier(model, binding);
        StringBuffer modifiers = GenerationUtil.getModifiers(modifier, false);
        
        String returnTypeName = importRewrite.addImport(binding.getReturnType());
		StringBuffer parameterList = GenerationUtil.getParametersList(binding, (IMethod)binding.getJavaElement(), importRewrite);
		StringBuffer exceptionList = GenerationUtil.getExceptionsList(binding, importRewrite);

		content.append(modifiers);
        
        StringBuffer methodTypeParameterList = GenerationUtil.getTypeParameterList(binding, importRewrite);
        if (methodTypeParameterList.length() != 0) {
        	content.append("<").append(methodTypeParameterList).append("> ");
        }

        if (!binding.isConstructor()) {
        	content.append(returnTypeName).append(" ");
        } 
        content.append(methodName);
        	
        content.append("(").append(parameterList).append(")");
		if (exceptionList.length() != 0) {
			content.append(" throws ").append(exceptionList);
        }
		content.append(" {").append(delimiter).append("\t");
		
		if (!binding.isConstructor() && !binding.getReturnType().getName().equals("void")) {
			content.append("return ");
		}
		
		content.append(invocationPrefix);
		content.append("(");
		
        String[] parameterNames = ((IMethod)binding.getJavaElement()).getParameterNames();
        
        for (int i=0; i<parameterNames.length; i++) {
        	content.append(parameterNames[i]);
        	if (i < parameterNames.length - 1) {
        		content.append(", ");
        	}
        }
        content.append(");").append(delimiter).append("}");
        
        return modifier;
	}

	private int getModifier(InferTypeConstraintsModel model, IMethodBinding binding) {		
		int modifiers = binding.getModifiers();
		if (model != null && (Modifier.isProtected(modifiers) || Modifier.isPublic(modifiers))) {
			if (newSuperclass == null || Bindings.findMethodInHierarchy(newSuperclass.getBaseTypeBinding(), binding.getName(), binding.getMethodDeclaration().getParameterTypes()) == null) {
				boolean hasExternalUse = false;
				boolean hasSubtypeUse = false;
				
		    	Set<TType> subTypes = new HashSet<TType>();
		    	for (ExistingTCType existingTCType : typeHierarchy.getSubTypes()) {
					subTypes.add(model.createTType(existingTCType.getBaseTypeBinding()));
				}
				String key = binding.getMethodDeclaration().getKey();

				for (ITypeConstraint2 constraint : model.getTypeConstraints()) {
					if (constraint.getRight() instanceof BindingKeyProvider) {							
						String bindingKey = ((BindingKeyProvider)constraint.getRight()).getBindingKey();
						if (key.equals(bindingKey)) {
							if (constraint.getLeft() instanceof ThisVariable || constraint.getLeft() instanceof SuperVariable) {
								TType leftType = constraint.getLeft().getType();
								if (leftType != null && subTypes.contains(leftType.getErasure())) {
									hasSubtypeUse = true;
								} 
							} else {
								hasExternalUse = true;									
							}
						}							
					}
				}
				
				if (!hasExternalUse) {
					if (hasSubtypeUse) {
						modifiers ^= modifiers & (Modifier.PRIVATE | Modifier.PUBLIC);
						modifiers |= Modifier.PROTECTED;
			        } else {
						modifiers ^= modifiers & (Modifier.PROTECTED | Modifier.PUBLIC);
						modifiers |= Modifier.PRIVATE;
			        }
				}
			}
		}
		
		return modifiers;
	}

	private void createField(ExistingTCType type,
			final AbstractTypeDeclaration declaration)
			throws JavaModelException, InternalCalculationException {
		ASTRewrite astRewrite = sourceRewrite.getASTRewrite();
		ImportRewrite importRewrite = sourceRewrite.getImportRewrite();
				
		StringBuilder builder = new StringBuilder();
		builder.append("protected final ").append(delegateeClassName);
		addTypeArguments(builder, type, importRewrite);
		
		builder.append(" ").append(delegateeFieldName);

        if (!hasConstructors(type)) {
	        builder.append(" = new ").append(delegateeClassName);
			addTypeArguments(builder, type, importRewrite);
	        builder.append("()");
		}
		builder.append(";").append(StubUtility.getLineDelimiterUsed(type.getIType()));
		
		ASTNode delegateeField = astRewrite.createStringPlaceholder(builder.toString(), ASTNode.FIELD_DECLARATION);
		ListRewrite listRewriter = astRewrite.getListRewrite(declaration, declaration.getBodyDeclarationsProperty());
		listRewriter.insertFirst(delegateeField, sourceRewrite.createCategorizedGroupDescription("Add delegatee field", SET_RIWD));
	}

    private boolean hasConstructors(ExistingTCType type) throws JavaModelException {
    	for (IMethod method : type.getIType().getMethods()) {
			if (method.isConstructor()) {
				return true;
			}
		}
		return false;
	}

	@SuppressWarnings("unchecked")
    private void changeTypeSignature(final AbstractTypeDeclaration declaration, Set<TType> requiredInterfaces, final RefactoringStatus status)
            throws JavaModelException, InternalCalculationException {
        Assert.isNotNull(declaration);
        Assert.isNotNull(status);
        
        ExistingTCType modifiedType = typeHierarchy.getOriginalSelectedType();
        IType subType = modifiedType.getIType();

        final AST ast = declaration.getAST();
        final ITypeParameter[] parameters = subType.getTypeParameters();

        final ASTRewrite rewriter = sourceRewrite.getASTRewrite();

        ListRewrite listRewrite = null;        
        if (declaration instanceof TypeDeclaration) {
            listRewrite = rewriter.getListRewrite(declaration, TypeDeclaration.SUPER_INTERFACE_TYPES_PROPERTY);
        } else if (declaration instanceof EnumDeclaration) {
            listRewrite = rewriter.getListRewrite(declaration, EnumDeclaration.SUPER_INTERFACE_TYPES_PROPERTY);
        }
        
        
        if (listRewrite != null) { 
            CategorizedTextEditGroup descr = sourceRewrite.createCategorizedGroupDescription(RefactoringCoreMessages.ExtractInterfaceProcessor_add_super_interface, SET_RIWD);
            for (TType newType : requiredInterfaces) {
                Type type = getNewSuperType(ast, parameters, typeHierarchy.getType(newType), modifiedType);
                listRewrite.insertLast(type, descr);
            }
        }
        
        if (newSuperclass != null && !newSuperclass.isObject()) {
            Type type = getNewSuperType(ast, parameters, newSuperclass, modifiedType);
            rewriter.set(declaration, TypeDeclaration.SUPERCLASS_TYPE_PROPERTY, type, sourceRewrite.createCategorizedGroupDescription("Change super type", SET_RIWD));
        } else {
            rewriter.set(declaration, TypeDeclaration.SUPERCLASS_TYPE_PROPERTY, null, sourceRewrite.createCategorizedGroupDescription("Change super type", SET_RIWD));        	
        }
        
    }

    @SuppressWarnings("unchecked")
    private Type getNewSuperType(final AST ast, final ITypeParameter[] parameters, TCType newType, ExistingTCType modifiedType) throws JavaModelException, InternalCalculationException {
        String completeTypeSignature = newType.getHierarchyTypeSignature(modifiedType);
        if (completeTypeSignature == null) {
            Type type = ast.newSimpleType(ast.newSimpleName(sourceRewrite.getImportRewrite().addImport(newType.getTypeName())));
            if (parameters.length > 0) {
                final ParameterizedType parameterized = ast.newParameterizedType(type);
                for (ITypeParameter element : parameters) {
                    parameterized.typeArguments().add(ast.newSimpleType(ast.newSimpleName(element.getElementName())));
                }
                type = parameterized;
            }
            return type;
        } else {
            return sourceRewrite.getImportRewrite().addImportFromSignature(completeTypeSignature, sourceRewrite.getAST());
        }
    }
    
	public Stats getStats() {
		return stats;
	}    

	public static class Stats {
		
		private int nonPrivateMethodsBefore;
		private int nonPrivateMethodsAfter;
		private int nonPrivateStaticMethodsBefore;
		private int nonPrivateStaticMethodsAfter;
		private int delegatingMethods;
		private int reverseDelegatingMethods;
		private int superDelegatingMethods;
		private int newInterfaces;
		private int deadMethods;
		private boolean newSupertype;
		
		private int newlyDeclaredNonPrivateMethodsBefore;
		private int newlyDeclaredNonPrivateStaticMethodsBefore;
		private int onlyInheritedMethodsBefore;
		private int onlyInheritedStaticMethodsBefore;
		
		public int getNonPrivateMethodsBefore() {
			return nonPrivateMethodsBefore;
		}
		public int getNonPrivateMethodsAfter() {
			return nonPrivateMethodsAfter;
		}
		public int getNonPrivateStaticMethodsBefore() {
			return nonPrivateStaticMethodsBefore;
		}
		public int getNonPrivateStaticMethodsAfter() {
			return nonPrivateStaticMethodsAfter;
		}
		public int getDelegatingMethods() {
			return delegatingMethods;
		}
		public int getReverseDelegatingMethods() {
			return reverseDelegatingMethods;
		}
		public int getSuperDelegatingMethods() {
			return superDelegatingMethods;
		}
		public int getNewInterfaces() {
			return newInterfaces;
		}
		public int getDeadMethods() {
			return deadMethods;
		}		
		public boolean isNewSupertype() {
			return newSupertype;
		}
		public int getNewlyDeclaredNonPrivateMethodsBefore() {
			return newlyDeclaredNonPrivateMethodsBefore;
		}
		public int getNewlyDeclaredNonPrivateStaticMethodsBefore() {
			return newlyDeclaredNonPrivateStaticMethodsBefore;
		}
		public int getOnlyInheritedMethodsBefore() {
			return onlyInheritedMethodsBefore;
		}
		public int getOnlyInheritedStaticMethodsBefore() {
			return onlyInheritedStaticMethodsBefore;
		}
		
	}	
}
