/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoj.riwd.ui;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.viewers.CheckboxTreeViewer;
import org.eclipse.ltk.ui.refactoring.UserInputWizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Text;
import org.intoJ.inferType3.InferTypePlugin;
import org.intoJ.inferType3.InferTypeRefactoring;
import org.intoJ.inferType3.internal.typeCalc.AccessSet;
import org.intoj.riwd.RiwdRefactoring;

/**
 * The single custom wizard page of the refactoring wizard.
 * 
 * @author Hannes Kegel
 *
 */
class RiwdInputPage extends UserInputWizardPage {

    private static final int WIDTH = 350;
    private static final String PAGE_NAME= "InferTypeInputPage";

    private final RiwdRefactoring refactoring;

    private boolean initialized = false;
    private Composite newComposite;
    		
	/**
     * The single constructor.
     * 
	 * @param inferTypeWizard the wizard
	 */
	public RiwdInputPage(RiwdWizard inferTypeWizard) {
		super(PAGE_NAME);
        refactoring = (RiwdRefactoring)inferTypeWizard.getRefactoring();
	}

    /* (non-Javadoc)
     * @see org.eclipse.ltk.ui.refactoring.UserInputWizardPage#setVisible(boolean)
     */
    @Override
    public void setVisible(boolean visible) {
        if (visible && !initialized) {
            initialized = true;
//            try {
                final RiwdRefactoring riwdRefactoring = ((RiwdRefactoring)getRefactoring());
                Composite composite = (Composite)getControl();
                addOptionGroup(composite);
                //createNewComposite(composite, riwdRefactoring);
                Dialog.applyDialogFont(composite);
                
//            } catch (JavaModelException e) {
//                e.printStackTrace();
//                setErrorMessage("Could not determine access sets");
//            }
        }
        super.setVisible(visible);
    }
    

    /* (non-Javadoc)
     * @see org.eclipse.jface.dialogs.IDialogPage#createControl(org.eclipse.swt.widgets.Composite)
     */
    public void createControl(Composite parent) {
        initializeDialogUnits(parent);
        Composite composite= new Composite(parent, SWT.NONE);
        setControl(composite);
        GridLayout gridLayout = new GridLayout(1, false);
        gridLayout.verticalSpacing = 8;
        composite.setLayout(gridLayout);
    }
    

    private void addOptionGroup(Composite composite) {
        Group group = new Group(composite, SWT.NONE);
        group.setText("Options");
        GridData gridData = new GridData(SWT.FILL, SWT.FILL, true, false);
        gridData.verticalIndent = 8; 
        group.setLayoutData(gridData);
        
        GridLayout gridLayout2 = new GridLayout(1, false);
        gridLayout2.verticalSpacing = 8;
        gridLayout2.marginTop = 3;
        group.setLayout(gridLayout2);

        final RiwdRefactoring riwdRefactoring = ((RiwdRefactoring)getRefactoring());

        final Button forwardingButton = new Button(group, SWT.RADIO);
        gridData = new GridData(SWT.LEFT, SWT.TOP, true, false);
        forwardingButton.setLayoutData(gridData);
        forwardingButton.setText(" Forwarding");
		forwardingButton.setSelection(riwdRefactoring.isPerformForwarding());
        forwardingButton.setEnabled(riwdRefactoring.getForwardingStatus().isPossible());
        forwardingButton.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                riwdRefactoring.setPerformForwarding(true);
            }
        });

        final Button delegationButton = new Button(group, SWT.RADIO);
        gridData = new GridData(SWT.LEFT, SWT.TOP, true, false);
        delegationButton.setLayoutData(gridData);
        delegationButton.setText(" Delegation");
        delegationButton.setSelection(!riwdRefactoring.isPerformForwarding());
        delegationButton.setEnabled(riwdRefactoring.getDelegationStatus().isPossible());
        delegationButton.addSelectionListener(new SelectionAdapter() {
            @Override
            public void widgetSelected(SelectionEvent e) {
                riwdRefactoring.setPerformForwarding(false);
            }
        });
        
        if (!riwdRefactoring.getForwardingStatus().isPossible()) {
            Label label = new Label(group, SWT.WRAP);
            gridData = new GridData(SWT.FILL, SWT.TOP, true, false, 2, 1);
            gridData.widthHint = WIDTH;
            label.setLayoutData(gridData);
            label.setText("Forwarding not possible:\n" + riwdRefactoring.getForwardingStatus().getViolationDescription());        	
        }
        
        if (!riwdRefactoring.getDelegationStatus().isPossible()) {
            Label label = new Label(group, SWT.WRAP);
            gridData = new GridData(SWT.FILL, SWT.TOP, true, false, 2, 1);
            gridData.widthHint = WIDTH;
            label.setLayoutData(gridData);
            label.setText("Delegation not possible:\n" + riwdRefactoring.getDelegationStatus().getViolationDescription());        	
        }

    }

    private void createNewComposite(Composite composite, final RiwdRefactoring riwdProcessor) throws JavaModelException {
        newComposite = new Composite(composite, SWT.NONE);
        GridData gridData = new GridData(SWT.FILL, SWT.FILL, true, true);
        newComposite.setLayoutData(gridData);

        GridLayout gridLayout2 = new GridLayout(2, false);
        gridLayout2.verticalSpacing = 8;
        gridLayout2.marginHeight = 0;
        gridLayout2.marginWidth = 0;
        gridLayout2.marginRight= 0;
        gridLayout2.marginTop = 0;
        gridLayout2.marginBottom = 0;
        newComposite.setLayout(gridLayout2);

        Label label = new Label(newComposite, SWT.WRAP);
        gridData = new GridData(SWT.FILL, SWT.TOP, true, false, 2, 1);
        gridData.widthHint = WIDTH;
        label.setLayoutData(gridData);
        label.setText("The initially selected methods are used on the reference and must be contained in the new type. " + 
                "You can add additional methods to the new type below.");
                    
        CheckboxTreeViewer checkBoxTreeViewer = new CheckboxTreeViewer(newComposite);

//        AccessSet accessSet = riwdProcessor.getAccessSet();            
//        ITContentProvider contentProvider = new ITContentProvider(accessSet);
//        checkBoxTreeViewer.setContentProvider(contentProvider);
//        checkBoxTreeViewer.setLabelProvider(new ITLabelProvider());
        gridData = new GridData(SWT.FILL, SWT.FILL, true, true, 2, 1);
        gridData.heightHint = 250;
        gridData.widthHint = WIDTH;
        checkBoxTreeViewer.getControl().setLayoutData(gridData);
//        checkBoxTreeViewer.setInput(riwdProcessor.getTypeHierarchy());
//        checkBoxTreeViewer.setCheckedElements(accessSet.getMembers().toArray());
//        checkBoxTreeViewer.addCheckStateListener(new CheckStateListener(refactoring));
//        checkBoxTreeViewer.expandAll();
//        Object[] elements = ((ITContentProvider)checkBoxTreeViewer.getContentProvider()).getElements(riwdProcessor.getTypeHierarchy());
//        if (elements.length > 0) {
//            checkBoxTreeViewer.reveal(elements[0]);
//        }
        
   }
   
}